# Copyright (C) 2015 Red Hat
#
# This file is part of testdays.
#
# testdays is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""CLI module for testdays."""

import argparse
import bugzilla
import datetime
import sys

import wikitcms.page
import wikitcms.release
import wikitcms.result
import wikitcms.wiki

def datecheck(string):
    if not string:
        return None
    try:
        date = datetime.datetime.strptime(string, '%Y%m%d')
    except ValueError:
        try:
            date = datetime.datetime.strptime(string, '%Y-%m-%d')
        except ValueError:
            raise ValueError('Date must be of form 20141208 or 2014-12-08')
    return date.strftime('%Y-%m-%d')

class Cli(object):
    """Bits for the testdays CLI tool."""
    def __init__(self):
        self.args = self.parse_args()
        # We don't do anything that needs authentication yet, so just
        # instantiate the wiki and bugzilla but don't log in
        self.site = wikitcms.wiki.Wiki()
        self.bz = bugzilla.RHBugzilla(
            url='https://bugzilla.redhat.com/xmlrpc.cgi')

    def get_stats(self, testday):
        """Get statistics from a single test day page."""
        testers = dict()
        openbugs = fixedbugs = dupebugs = unfixedbugs = []
        # Extract all the results from the page
        results = wikitcms.result.find_results_by_row(
            testday.results_wikitext, statuses=('pass', 'fail', 'warn'))
        for res in results:
            # Build a dict keyed on username, counting number of tests
            # that user performed and building a set of bugs referred
            # to by that user (so we can count unique bug references)
            if res.user in testers:
                testers[res.user][0] += 1
                testers[res.user][1].update(res.bugs)
            else:
                testers[res.user] = [1, set(res.bugs)]

        # Generate bug stats using the 'bugs' page property, which
        # also finds bugs that can't be associated with a specific
        # result
        bugs = self.bz.getbugs(testday.bugs)
        # Filter out Nones caused by bad bug IDs
        bugs = [bug for bug in bugs if bug]
        if bugs:
            # We 'bin' the bugs in order to try and derive a useful
            # number for 'percentage of valid, unique bugs fixed'.
            # For more details see:
            # https://lists.fedoraproject.org/pipermail/test/2010-April/090271.html
            openbugs = [bug for bug in bugs if bug.status != 'CLOSED']
            fixedbugs = [bug for bug in bugs if
                         bug.resolution in ('ERRATA', 'RAWHIDE',
                                            'CURRENTRELEASE', 'NEXTRELEASE')]
            dupebugs = [bug for bug in bugs if
                        bug.resolution in ('DUPLICATE', 'NOTABUG',
                                           'WORKSFORME', 'UPSTREAM')]
            unfixedbugs = [bug for bug in bugs if
                           bug.resolution in ('CANTFIX', 'WONTFIX', 'DEFERRED',
                                              'INSUFFICIENT_DATA', 'EOL')]
        # Return all the obvious things, plus a count of all results,
        # to save us laboriously reconstructing it from the dict later
        return (testers, str(len(results)), openbugs, fixedbugs, dupebugs,
                unfixedbugs)

    def print_stats(self, testdays):
        """Output statistics for one or more test days."""
        # These vars just count up some particular values from *all*
        # chosen test days, for the overall stats at the end.
        allfixed = set()
        allbugs = set()
        alldupe = set()
        allresults = 0
        alltesters = dict()

        for testday in testdays:
            print(testday.name)
            # Get the raw data
            (testers, results, openbugs, fixedbugs,
             dupebugs, unfixedbugs) = self.get_stats(testday)
            # Update the 'all' counters
            allresults += int(results)
            allfixed.update(bug.id for bug in fixedbugs if bug)
            alldupe.update(bug.id for bug in dupebugs if bug)
            for tester in testers:
                if tester in alltesters:
                    alltesters[tester][0] += testers[tester][0]
                    alltesters[tester][1].update(testers[tester][1])
                else:
                    alltesters[tester] = testers[tester]

            bugs = openbugs + fixedbugs + dupebugs + unfixedbugs
            allbugs.update(bugs)
            if self.args.list_bugs:
                # Print the list of one-line bug summaries
                print("")
                bugs.sort(key=lambda x: x.id)
                for bug in bugs:
                    print(bug)
                print("")

            # Sometimes we can't calculate these, so set a fallback
            fixedpercent = ratio = "N/A"
            if bugs and testers:
                # This is just 'bugs submitted per tester'
                ratio = float(len(bugs)) / len(testers)
            if len(bugs) - len(dupebugs) > 0:
                # This is the 'percentage of bugs fixed' calculation
                fixedpercent = float(
                    len(fixedbugs)) / (len(bugs) - len(dupebugs)) * 100
            print("Testers: {0}, Tests: {1}, Bugs: {2}, Ratio: {3}".format(
                str(len(testers)), results, str(len(bugs)), str(ratio)))
            print("Open: {0}, Dupe: {1}, Fixed: {2}, Unfixed: {3}, "
                  "Fixed %: {4}\n".format(
                    str(len(openbugs)), str(len(dupebugs)),
                    str(len(fixedbugs)), str(len(unfixedbugs)),
                    str(fixedpercent)))

        if len(testdays) > 1:
            print("Events: {0}\tTests: {1}\tBugs: {2}\tTesters: {3}\t".format(
                str(len(testdays)), str(allresults), str(len(allbugs)),
                str(len(alltesters))))
            # This is the 'percentage fixed' calculation again, this
            # time across all chosen test days
            if len(allbugs) - len(alldupe) > 0:
                print("Overall fixed %: {0}\n".format(
                    str(float(len(allfixed)) / (len(allbugs) - len(alldupe)) * 100)))

            # To find the top 10 testers we sort the alltesters dict
            # using the first element of the 'value' tuple, which is
            # the count of tests for that tester.
            toptesters = sorted(list(alltesters.keys()),
                                key=lambda x: alltesters[x][0], reverse=True)
            # Now we find the top 10 bug referrers.
            topreferrers = sorted(list(alltesters.keys()),
                                  key=lambda x: len(alltesters[x][1]),
                                  reverse=True)
            print("Top Testers")
            # Start with the top ten testers by test count...
            tops = toptesters[0:10]
            # ...now add in the top ten by bug referral count, so long
            # as their count is at least 2.
            for tester in topreferrers[0:10]:
                if tester not in tops and len(alltesters[tester][1]) > 1:
                    tops.append(tester)
            # For each tester in the combined list, print their test
            # count and list of bugs referred.
            for tester in tops:
                print("Tester: {0} Tests: {1} Bugs: {2}".format(
                    tester, str(alltesters[tester][0]),
                    ', '.join(alltesters[tester][1])))

    def stats(self):
        """stats sub-command method. Prints statistics for the
        Test Day(s) specified on the command line."""
        if self.args.release and (
            self.args.until or (self.args.since and
                                self.args.since != '2010-09-01')):
            print("WARNING: --since and/or --until specified along with "
                  "--release. --since and --until will be ignored.")
        if self.args.release:
            rel = wikitcms.release.Release(self.args.release, self.site)
            testdays = rel.testday_pages
        else:
            # Use the 'alltestdays' generator and specify 'start' and
            # 'end' values if we were given since/until dates
            testdays = self.site.alltestdays(
                start=self.args.since, end=self.args.until,
                redirects='nonredirects')
            # 'alltestdays' doesn't check the page is actually a real
            # Test Day, it can include some stray things like subpages
            # - throw those out.
            testdays = [t for t in testdays if
                        isinstance(t, wikitcms.page.TestDayPage)]
        if self.args.filter:
            testdays = [t for t in testdays if
                        self.args.filter.lower() in t.name.lower()]
        # We're generally going to want to see the pages in date
        # order.
        testdays.sort(key=lambda x:x.name)
        self.print_stats(testdays)

    def modify(self):
        """modify sub-command method. Makes changes to the Test Day
        specified on the command line. Requires credentials file.
        """
        self.site.login()
        page = self.site.pages[self.args.name]
        if not page.exists:
            sys.exit("Page {0} does not exist!".format(self.args.name))
        if self.args.results:
            print("Fixing results...")
            page.fix_app_results()
        if self.args.longrefs:
            print("Fixing long refs...")
            page.long_refs()
        sys.exit()

    def parse_args(self):
        """Parse arguments with argparse."""
        parser = argparse.ArgumentParser(description=(
            "Tool for interacting with Test Days in the Fedora wiki. "))
        # This is a workaround for a somewhat infamous argparse bug
        # in Python 3. See:
        # https://stackoverflow.com/questions/23349349/argparse-with-required-subparser
        # http://bugs.python.org/issue16308
        subparsers = parser.add_subparsers(dest='stats')
        subparsers.required = True

        parser_stats = subparsers.add_parser(
            'stats', description="Print statistics about one or more Test "
            "Days.")
        parser_stats.add_argument(
            '-r', '--release', help="Only process Test Days from this release."
            " Can be combined with --filter. Cannot be combined with --since "
            "and/or --until: if both are specified, --since and --until will "
            "be ignored. Note that Test Days from Fedora 13 and earlier often "
            "cannot be parsed correctly and may well give unreliable data.",
            type=int, required=False, choices=list(range(1, 100)), metavar="1-99")
        parser_stats.add_argument(
            '-f', '--filter', help="Only process Test Days in whose name this "
            "string appears (case-insensitive). Can be combined with --release"
            "OR --since / --until.", required=False)
        parser_stats.add_argument(
            '-s', '--since', help="Only process Test Days from this date "
            "onwards. Can be combined with --until and/or --filter. Defaults "
            "to 2010-09-01, because Test Days earlier than that date usually "
            "cannot be parsed correctly and will give unreliable data.",
            default='2010-09-01', type=datecheck, required=False)
        parser_stats.add_argument(
            '-u', '--until', help="Only process Test Days from before this "
            "date. Can be combined with --since and/or --filter.",
            default=None, type=datecheck, required=False)
        parser_stats.add_argument(
            '-l', '--list-bugs', help="Print a one-line summary of each bug "
            "report for each Test Day.", required=False, action='store_true')
        parser_stats.set_defaults(func=self.stats)

        parser_modify = subparsers.add_parser(
            'modify', description="Modify Test Day pages.")
        parser_modify.add_argument(
            'name', help="Name of Test Day page to modify.")
        parser_modify.add_argument(
            '-r', '--results', help="Fix {{result}} template usage.",
            action='store_true')
        parser_modify.add_argument(
            '-l', '--longrefs', help="Fix long references.",
            action='store_true')
        parser_modify.set_defaults(func=self.modify)

        args = parser.parse_args()
        return args

    def run(self):
        """Read in arguments and run sub-command function."""
        self.args.func()

def main():
    """Main loop."""
    try:
        cli = Cli()
        cli.run()
    except KeyboardInterrupt:
        sys.stderr.write("Interrupted, exiting...\n")
        sys.exit(1)

if __name__ == '__main__':
    main()
